from lxml import etree
from lepl.apps.rfc3696 import HttpUrl

from StringIO import StringIO
import json
import urllib2
import sys
import os.path
from HTMLParser import HTMLParser, HTMLParseError

sys.path.append(
    os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir)))

from errors import _error
from validationresult import ValidationResult, ValidationWarning
from validator import SchemaValidator

PARSELY_PAGE_SCHEMA = "http://parsely.com/static/data/parsely_page_schema.html"


class ParselyPageParser(HTMLParser):
    def __init__(self):
        HTMLParser.__init__(self)
        self.ppage = None

    def handle_starttag(self, tag, attrs):
        if self.ppage is not None or tag != "meta":
            return

        attrs = dict(attrs)
        if attrs.get("name") == "parsely-page":
            ppage = attrs.get("content", attrs.get("value"))
            if ppage:
                try:
                    self.ppage = json.loads(ppage)
                except:
                    raise HTMLParseError("bad ppage")  # bad ppage

    def original_unescape(self, s):
        """Since we need to use this sometimes"""
        if isinstance(s, basestring):
            return unicode(HTMLParser.unescape(self, s))
        elif isinstance(s, list):
            return [unicode(HTMLParser.unescape(self, item)) for item in s]
        else:
            return s

    def unescape(self, s):
        return s


class ParselyPageValidator(SchemaValidator):
    def __init__(self, graph, doc_lines, url=""):
        super(ParselyPageValidator, self).__init__(graph, doc_lines, url=url)
        self.text = '\n'.join([a[0] for a in doc_lines])
        self.stdref = self.get_standard()
        self.url_validator = HttpUrl()
        self.data = self._get_parselypage(self.text)

    def get_standard(self):
        """get list of allowed parameters"""
        try:
            res = urllib2.urlopen(PARSELY_PAGE_SCHEMA)
        except:
            return []
        text = res.read()
        tree = etree.parse(StringIO(text))
        stdref = tree.xpath("//div/@about")
        return [a.split(':')[1] for a in stdref]

    def _get_parselypage(self, body):
        """extract the parsely-page meta content from a page"""
        parser = ParselyPageParser()
        ret = None
        try:
            parser.feed(body)
        except HTMLParseError:
            pass  # ignore and hope we got ppage
        if parser.ppage is None:
            return

        ret = parser.ppage
        if ret:
            ret = {parser.original_unescape(k): parser.original_unescape(v)
                   for k, v in ret.iteritems()}
        return ret

    def validate(self):
        result = ValidationResult("parsely-page", self.__class__.__name__)

        if self.data:
            for key in self.data.keys():
                res = self.check_key(key)
                if res:
                    result.add_error(res)

        return result

    def check_key(self, key):
        if key not in self.stdref:
            err = _error("{0} - invalid parsely-page field", key,
                         doc_lines=self.doc_lines)
            return ValidationWarning(ValidationResult.ERROR, err['err'], err['line'], err['num'])
        if key in ["link", "image_url"]:
            if not self.url_validator(self.data[key]):
                err = _error("{0} - invalid url for field '{1}'", self.data[key], key,
                             doc_lines=self.doc_lines)
                return ValidationWarning(ValidationResult.ERROR, err['err'], err['line'], err['num'])
        return None



from distutils.core import setup

setup(
    name='schemato',
    version='1.0',
    py_modules=['schemato'],

    author='emmett butler',
    author_email='emmett.butler321@gmail.com',
    url='http://github.com/emmett9001/schemato',
    description='Unified semantic metadata validator'
)



from rdflib.plugins.parsers.pyMicrodata import pyMicrodata
from rdflib.plugins.parsers.pyRdfa import pyRdfa


class CompoundGraph(object):
    def __init__(self, source):
        super(CompoundGraph, self).__init__()
        try:
            self.microdata_graph = pyMicrodata().graph_from_source(source)
        except:
            self.microdata_graph = None

        try:
            self.rdfa_graph = pyRdfa().graph_from_source(source)
        except:
            self.rdfa_graph = None



from distillery import ParselyDistiller, NewsDistiller
from schemato import Schemato
from pprint import pprint
import unittest


class TestSchemato(unittest.TestCase):
    def test_schema_no_errors(self):
        scm = Schemato("test_documents/schema.html")
        results = scm.validate()
        for res in results:
            self.assertTrue(len(res) == 0)

    def test_schema_errors(self):
        scm = Schemato("test_documents/schema_errors.html")
        results = scm.validate()

        expected = {'classname': 'SchemaOrgSchemaDef',
                    'errors': [
                        {'line': '&lt;a  itemprop="copyrightNotice"',
                         'num': 63,
                         'string': 'copyrightNotice - invalid member of NewsArticle',
                         'level': 'Error'},
                        {'line': '&lt;meta itemprop="tickerSymbol" content="NYSE NYT"/&gt;',
                         'num': 74,
                         'string': 'tickerSymbol - invalid member of Organization',
                         'level': 'Error'},
                        {'line': 'itemprop="createdBy"',
                         'num': 47,
                         'string': 'createdBy - invalid member of NewsArticle',
                         'level': 'Error'},
                        {'line': '&lt;a  itemprop="usageTerms"',
                         'num': 78,
                         'string': 'usageTerms - invalid member of NewsArticle',
                         'level': 'Error'},
                        {'line': 'itemtype="http://schema.org/UserComment"',
                         'num': 111,
                         'string': 'UserComment - invalid class',
                         'level': 'Error'},
                        {'line': 'itemtype="http://schema.org/UserComment"',
                         'num': 111,
                         'string': 'UserComment - invalid class',
                         'level': 'Error'},
                        {'line': 'itemtype="http://schema.org/UserComment"',
                         'num': 111,
                         'string': 'UserComment - invalid class',
                         'level': 'Error'}
                    ],
                    'namespace': 'http://schema.org/',
                    'warnings': []}

        for res in results:
            if res.classname == 'SchemaOrgSchemaDef':
                self.assertTrue(len(res.errors) > 0)
                for err in res.errors:
                    self.assertTrue(err.string in [a['string'] for a in expected['errors']])

    def test_html_parse(self):
        text = open("test_documents/foxnews.html", "r").read()
        scm = Schemato(text, url="http://foxnews.com", loglevel="ERROR")
        results = scm.validate()
        for res in results:
            if res.classname == "ParselyPageValidator":
                self.assertTrue("ttle - invalid parsely-page field" in [a.string for a in res.errors])
                self.assertTrue(len(res.errors) == 1)

    """
    def test_distillers(self):
        print "Loading NY Daily News..."
        nydailynews = Schemato("http://www.nydailynews.com/news/politics/obama-fights-back-2nd-debate-romney-article-1.1185271")
        print "Done."
        print "Loading Mashable..."
        mashable = Schemato("http://mashable.com/2012/10/17/iphone-5-supply-problems/")
        print "Done."

        def demo(desc, class_, site):
            print desc
            print "=" * len(desc)
            d = class_(site)
            d.distill()
            pprint({"distilled": d.distilled, "sources": d.sources})

        #demo("Parse.ly strategy on Mashable", ParselyDistiller, mashable)
        demo("News strategy on Mashable", NewsDistiller, mashable)
        #demo("Parse.ly strategy on NY Daily News", ParselyDistiller, nydailynews)
        demo("News strategy on NY Daily News", NewsDistiller, nydailynews)
    """


if __name__ == "__main__":
    unittest.main()



VALIDATOR_MODULES = [
    "schemas.rnews.RNewsValidator",
    "schemas.opengraph.OpenGraphValidator",
    "schemas.schemaorg.SchemaOrgValidator",
    "schemas.schemaorg_rdf.SchemaOrgRDFaValidator",
    "schemas.parselypage.ParselyPageValidator",
]

# root of schema cache
CACHE_ROOT = "/tmp"
# how many seconds to wait until re-cache
CACHE_EXPIRY = 60 * 60 * 500



from validator import RdfValidator
from schemadef import RdfSchemaDef


class SchemaOrgRDFaSchemaDef(RdfSchemaDef):
    def __init__(self):
        super(SchemaOrgRDFaSchemaDef, self).__init__()
        self._ontology_file = "http://schema.org/docs/schema_org_rdfa.html"
        self._representation = "schemaorg_schemadef"  # SAME AS MICRODATA VERSION


class SchemaOrgRDFaValidator(RdfValidator):
    def __init__(self, graph, doc_lines, url=""):
        super(SchemaOrgRDFaValidator, self).__init__(graph, doc_lines, url=url)
        self.schema_def = SchemaOrgRDFaSchemaDef()
        self.allowed_namespaces = ["http://schema.org/"]



from collections import defaultdict
import logging as log
import rdflib.term as rt

from validator import RdfValidator
from schemadef import RdfSchemaDef
from utils import deepest_node

class OpenGraphSchemaDef(RdfSchemaDef):
    def __init__(self, url):
        super(OpenGraphSchemaDef, self).__init__()
        self.source = url
        self._ontology_file = "http://ogp.me/ns/ogp.me.ttl"
        # representation is the filename of the cached local schema
        self._representation = "og_schemadef"
        self.parse_ontology()

    def parse_ontology(self):
        self.attributes_by_class = defaultdict(list, self.attributes_by_class)
        for subj, pred, obj in self._schema_nodes():
            leaves = [(subj, pred, obj)]
            if type(obj) == rt.BNode:
                leaves = deepest_node((subj, pred, obj), self.graph)

            for s,p,o in leaves:
                if s not in self.attributes_by_class[self.source] and not isinstance(s, rt.BNode):
                    self.attributes_by_class[self.source].append(s)

class OpenGraphValidator(RdfValidator):
    def __init__(self, graph, doc_lines, url=""):
        super(OpenGraphValidator, self).__init__(graph, doc_lines)
        self.source = url
        self.schema_def = OpenGraphSchemaDef(url)
        self.allowed_namespaces = ["http://ogp.me/ns#", "http://opengraphprotocol.org/schema/"]

    def _is_instance(self, (subj, pred, obj)):
        """helper, returns the class type of subj"""
        return self.source



import rdflib.term as rt

def deepest_node((subj, pred, obj), graph):
    """recurse down the tree and return a list of the most deeply nested
    child nodes of the given triple"""
    # i don't fully accept the premise that this docstring presents
    # i'm not a docstring literalist
    to_return = []
    def _deepest_node((subj, pred, obj), graph):
        children = []
        if isinstance(obj, rt.BNode):
            for s,p,o in graph:
                if str(s) == str(obj):
                    children.append((s,p,o))
            for s,p,o in children:
                s1,p1,o1 = _deepest_node((s, p, o), graph)
                # coupling *smacks hand with ruler*
                if "rNews" in str(o1) and (s1,p1,o1) not in to_return:
                    to_return.append((s1,p1,o1))
            return (s1,p1,o1)
        else:
            return (subj, pred, obj)
    _deepest_node((subj, pred, obj), graph)
    return to_return




from functools import partial

import rdflib.term as rt
import logging as log

from rdflib.plugins.parsers.pyMicrodata import pyMicrodata
from rdflib.plugins.parsers.pyRdfa import pyRdfa

from errors import _error
from validationresult import ValidationResult, ValidationWarning


class SchemaValidator(object):
    """ASSUMPTIONS:
        This class knows about the file being validated, but it recieves that
        file as a graph and a doc_lines
        It does not perform any parsing logic on the file
        It recieves a "validatable" graph object and returns errors"""
    def __init__(self, graph, doc_lines, url=""):
        super(SchemaValidator, self).__init__()
        self.schema_def = None
        self.allowed_namespaces = []
        self.graph = graph
        self.checked_attributes = []
        self.stripped_attribute_names = {}

        log.info("init validator: %s" % self.__class__.__name__)
        self.doc_lines = doc_lines

        self.err = partial(_error, doc_lines=self.doc_lines)

    def validate(self):
        """Iterate over all triples in the graph and validate each one
            appropriately
        """
        log.info("{}\nValidating against {}"
                 .format("-" * 100, self.schema_def.__class__.__name__))

        if not self.schema_def:
            raise ValueError("No schema definition supplied.")

        self.checked_attributes = []

        # TODO - this should maybe choose the actually used namespace, not just
        # the first one in the list
        result = ValidationResult(self.allowed_namespaces[0],
                                  self.schema_def.__class__.__name__)
        for subject, predicate, object_ in self.graph:
            log.info("\nsubj: {subj}\npred: {pred}\n obj: {obj}"
                     .format(subj=subject, pred=predicate,
                             obj=object_.encode('utf-8')))
            result.add_error(self._check_triple((subject, predicate, object_)))
        return result

    def _should_ignore_predicate(self, predicate):
        ignored_predicates = ['type', 'item', 'first', 'rest']
        return self._field_name_from_uri(predicate) in ignored_predicates

    def _check_triple(self, (subj, pred, obj)):
        """compare triple to ontology, return error or None"""
        if self._should_ignore_predicate(pred):
            log.info("Ignoring triple with predicate '{}'"
                     .format(self._field_name_from_uri(pred)))
            return

        classes = []
        log.warning("Possible member %s found" % pred)

        pred = self._expand_qname(pred)

        if self._namespace_from_uri(pred) not in self.allowed_namespaces:
            log.info("Member %s does not use an allowed namespace", pred)
            return

        instanceof = self._is_instance((subj, pred, obj))
        if type(instanceof) == rt.URIRef:
            instanceof = self._expand_qname(instanceof)

        if hasattr(self.schema_def, "attributes_by_class") and \
           not self.schema_def.attributes_by_class:
            log.info("Parsed ontology not found. Parsing...")
            self.schema_def.parse_ontology()

        class_invalid = self._validate_class(instanceof)
        if class_invalid:
            log.warning("Invalid class %s" % instanceof)
            return class_invalid
        # TODO - the above sometimes fails when a single object has more than one
        # rdfa type (eg <span property="schema:creator rnews:creator" typeof="schema:Person rnews:Person">
        # Graph chooses the type in an arbitrary order, so it's unreliable
        # eg: http://semanticweb.com/the-impact-of-rdfa_b35003

        classes = self._superclasses_for_subject(self.graph, instanceof)
        classes.append(instanceof)

        member_invalid = self._validate_member(pred, classes, instanceof)
        if member_invalid:
            log.warning("Invalid member of class")
            return member_invalid

        dupe_invalid = self._validate_duplication((subj, pred), instanceof)
        if dupe_invalid:
            log.warning("Duplication found")
            return dupe_invalid

        # collect a list of checked attributes
        self.checked_attributes.append((subj, pred))
        log.warning("successfully validated triple, no errors")
        return

    def _validate_class(self, cl):
        """return error if class `cl` is not found in the ontology"""
        if cl not in self.schema_def.attributes_by_class.keys():
            search_string = self._build_search_string(cl)
            err = self.err("{0} - invalid class", self._field_name_from_uri(cl),
                           search_string=search_string)
            return ValidationWarning(ValidationResult.ERROR, err['err'],
                                     err['line'], err['num'])

    def _get_stripped_attributes(self, member, classes):
        stripped = []
        if member not in self.stripped_attribute_names:
            for cl in classes:
                stripped.extend(
                    [self._field_name_from_uri(attr)
                     for attr in self.schema_def.attributes_by_class[cl]])
            self.stripped_attribute_names[member] = stripped
        else:
            stripped = self.stripped_attribute_names[member]
        return stripped

    def _validate_member(self, member, classes, instanceof):
        """return error if `member` is not a member of any class in `classes`"""
        log.info("Validating member %s" % member)

        stripped = self._get_stripped_attributes(member, classes)
        if self._field_name_from_uri(member) in stripped:
            all_class_members = sum([self.schema_def.attributes_by_class[cl]
                                     for cl in classes], [])
            if member in all_class_members:
                return
            if self._namespace_from_uri(member) in self.allowed_namespaces:
                err = self.err("Unoficially allowed namespace {0}",
                               self._namespace_from_uri(member))
                return ValidationWarning(ValidationResult.WARNING, err['err'],
                                         err['line'], err['num'])
        else:
            err = self.err("{0} - invalid member of {1}",
                           self._field_name_from_uri(member),
                           self._field_name_from_uri(instanceof))
            return ValidationWarning(ValidationResult.ERROR, err['err'],
                                     err['line'], err['num'])

    def _validate_duplication(self, (subj, pred), cl):
        """returns error if we've already seen the member `pred` on `subj`"""
        log.info("Validating duplication of member %s" % pred)
        if (subj, pred) in self.checked_attributes:
            err = self.err("{0} - duplicated member of {1}",
                           self._field_name_from_uri(pred),
                           self._field_name_from_uri(cl))
            return ValidationWarning(ValidationResult.WARNING, err['err'],
                                     err['line'], err['num'])

    def _superclasses_for_subject(self, graph, typeof):
        """helper, returns a list of all superclasses of a given class"""
        # TODO - this might be replacing a fairly simple graph API query where
        # it doesn't need to
        classes = []
        superclass = typeof
        while True:
            found = False
            for p, o in self.schema_def.ontology[superclass]:
                if self.schema_def.lexicon['subclass'] == str(p):
                    found = True
                    classes.append(o)
                    superclass = o
            if not found:
                break
        return classes

    def _is_instance(self, (subj, pred, obj)):
        """helper, returns the class type of subj"""
        input_pred_ns = self._namespace_from_uri(self._expand_qname(pred))
        triples = self.graph.triples(
            (subj, rt.URIRef(self.schema_def.lexicon['type']), None)
        )
        if triples:
            for tr in triples:
                triple_obj_ns = self._namespace_from_uri(self._expand_qname(tr[2]))
                if input_pred_ns == triple_obj_ns:  # match namespaces
                    return tr[2]  # return the object

    def _field_name_from_uri(self, uri):
        """helper, returns the name of an attribute (without namespace prefix)"""
        # TODO - should use graph API
        uri = str(uri)
        parts = uri.split('#')
        if len(parts) == 1:
            return uri.split('/')[-1] or uri
        return parts[-1]

    def _namespace_from_uri(self, uri):
        """returns the expanded namespace prefix of a uri"""
        # TODO - this could be helped a bunch with proper use of the graph API
        # it seems a bit fragile to treat these as simple string-splits
        uri = str(uri)
        parts = uri.split('#')
        if len(parts) == 1:
            return "%s/" % '/'.join(uri.split('/')[:-1])
        return "%s#" % '#'.join(parts[:-1])

    def _expand_qname(self, qname):
        """expand a qualified name's namespace prefix to include the resolved
        namespace root url"""
        if type(qname) is not rt.URIRef:
            raise TypeError("Cannot expand qname of type {}, must be URIRef"
                            .format(type(qname)))
        for ns in self.graph.namespaces():
            if ns[0] == qname.split(':')[0]:
                return rt.URIRef("%s%s" % (ns[1], qname.split(':')[-1]))
        return qname

    def _build_search_string(self, uri):
        return self._field_name_from_uri(uri)


class RdfValidator(SchemaValidator):
    def __init__(self, graph, doc_lines, url=""):
        super(RdfValidator, self).__init__(graph, doc_lines, url=url)
        self.parser = pyRdfa()
        self.graph = self.graph.rdfa_graph  # use the rdfa half of the compound graph


class MicrodataValidator(SchemaValidator):
    def __init__(self, graph, doc_lines, url=""):
        super(MicrodataValidator, self).__init__(graph, doc_lines, url=url)
        self.parser = pyMicrodata()
        self.graph = self.graph.microdata_graph  # use the microdata half of the compound

    def _build_search_string(self, uri):
        return str(uri)



import rdflib.term as rt
import logging as log

from rdflib.plugins.parsers.pyRdfa import pyRdfa
import rdflib
import os
import time
import urllib2

from collections import defaultdict

from utils import deepest_node
import settings


# what functionality is common to every single schema def?
class SchemaDef(object):
    """class to handle the loading and caching of a standard
    ASSUMPTIONS:
        This class holds the definition of a standard
        It does not perform any validation against that standard
        This class knows nothing about the input file being validated
    """

    errorstring_base = "Are you calling parse_ontology from the base SchemaDef class?"

    def __init__(self):
        super(SchemaDef, self).__init__()
        self._ontology_file = ""
        self.ontology = defaultdict(list)
        self.attributes_by_class = {}
        self._ontology_parser_function = None
        self.lexicon = {}

    def _read_schema(self):
        """return the local filename of the definition file for this schema
        if not present or older than expiry, pull the latest version from
        the web at self._ontology_file"""
        cache_filename = os.path.join(settings.CACHE_ROOT, "%s.smt" % self._representation)
        log.info("Attempting to read local schema at {}".format(cache_filename))
        try:
            if time.time() - os.stat(cache_filename).st_mtime > settings.CACHE_EXPIRY:
                log.warning("Cache expired, re-pulling")
                self._pull_schema_definition(cache_filename)
        except OSError:
            log.warning("Local schema not found. Pulling from web.")
            self._pull_schema_definition(cache_filename)
        else:
            log.info("Success")

        return cache_filename

    def _pull_schema_definition(self, fname):
        """download an ontology definition from the web"""
        std_url = urllib2.urlopen(self._ontology_file)
        cached_std = open(fname, "w+")
        cached_std.write(std_url.read())
        cached_std.close()

    def parse_ontology(self):
        """place the ontology graph into a set of custom data structures
        for use by the validator"""
        start = time.clock()
        log.info("Parsing ontology file for {}".format(self.__class__.__name__))
        for subj, pred, obj in self._schema_nodes():
            if subj not in self.attributes_by_class.keys():
                if obj == rt.URIRef(self.lexicon['class']) and pred == rt.URIRef(self.lexicon['type']):
                    self.attributes_by_class[subj] = []

            leaves = [(subj, pred, obj)]
            if type(obj) == rt.BNode:
                leaves = deepest_node((subj, pred, obj), self.graph)

            for s, p, o in leaves:
                if o not in self.attributes_by_class.keys():
                    self.attributes_by_class[o] = []
                if pred == rt.URIRef(self.lexicon['domain']):
                    self.attributes_by_class[o].append(subj)
        log.info("Ontology parsing complete in {}".format((time.clock() - start) * 1000))

    def _schema_nodes(self):
        """parse self._ontology_file into a graph"""
        name, ext = os.path.splitext(self._ontology_file)
        if ext in ['.ttl']:
            self._ontology_parser_function = lambda s: rdflib.Graph().parse(s, format='n3')
        else:
            self._ontology_parser_function = lambda s: pyRdfa().graph_from_source(s)
        if not self._ontology_parser_function:
            raise ValueError("No function found to parse ontology. %s" % errorstring_base)
        if not self._ontology_file:
            raise ValueError("No ontology file specified. %s" % errorstring_base)
        if not self.lexicon:
            raise ValueError("No lexicon object assigned. %s" % errorstring_base)

        latest_file = self._read_schema()

        try:
            self.graph = self._ontology_parser_function(latest_file)
        except:
            raise IOError("Error parsing ontology at %s" % latest_file)

        for subj, pred, obj in self.graph:
            self.ontology[subj].append((pred, obj))
            yield (subj, pred, obj)


# Implementation-specific subclasses
class RdfSchemaDef(SchemaDef):
    def __init__(self):
        super(RdfSchemaDef, self).__init__()
        # TODO - certainly not the best way to do this, should probably use the
        # pyRdfa/pyMicrodata graph APIs to make this more robust
        self.lexicon = {
            'range': "http://www.w3.org/2000/01/rdf-schema#range",
            'domain': "http://www.w3.org/2000/01/rdf-schema#domain",
            'type': "http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
            'subclass': "http://www.w3.org/2000/01/rdf-schema#subClassOf",
            'class': "http://www.w3.org/2002/07/owl#Class"
        }


class MicrodataSchemaDef(SchemaDef):
    def __init__(self):
        super(MicrodataSchemaDef, self).__init__()
        # TODO - certainly not the best way to do this, should probably use the
        # pyRdfa/pyMicrodata graph APIs to make this more robust
        self.lexicon = {
            'range': "http://schema.org/rangeIncludes",
            'domain': "http://schema.org/domainIncludes",
            'type': "http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
            'subclass': "http://www.w3.org/2000/01/rdf-schema#subClassOf",
            'class': "http://www.w3.org/2000/01/rdf-schema#Class"
        }



import urllib
import logging as log
import re
import os

from compound_graph import CompoundGraph
from StringIO import StringIO

from schemas.parselypage import ParselyPageValidator
import settings


class Schemato(object):
    def __init__(self, source, url=None, loglevel="ERROR"):
        """init with a local filepath or a URI"""
        super(Schemato, self).__init__()
        self.set_loglevel(loglevel)

        graph_source = source
        if url is not None:
            graph_source = StringIO(source)

        self.graph = CompoundGraph(graph_source)

        def _read_stream(source):
            text, url = self._get_document(source)
            return url, self._document_lines(text)

        parsed_url, self.doc_lines = _read_stream(source)
        self.url = url
        if url is None:
            self.url = parsed_url

        self.parsely_page = ParselyPageValidator(self.graph, self.doc_lines).data

    def validate(self):
        self._load_validators()

        results = [v.validate() for v in self.validators]
        log.info("returned from validate() : %s", results)
        for res in results:
            log.info(res.to_json())
        return results

    def set_loglevel(self, loglevel):
        if hasattr(log, loglevel):
            log.basicConfig(level=getattr(log, loglevel))
        else:
            log.basicConfig(level=log.ERROR)
            log.error("Unrecognized loglevel %s, defaulting to ERROR", loglevel)

    def _load_validators(self):
        def import_module(name):
            m = __import__(name)
            for n in name.split(".")[1:]:
                m = getattr(m, n)
            return m
        # include the parent directory in the path, to allow relative imports of
        # modules from settings
        os.sys.path.append(os.path.dirname(os.path.abspath(__file__)))
        self.validators = set()
        for module_path in settings.VALIDATOR_MODULES:
            path_parts = module_path.split('.')
            module = import_module(".".join(path_parts[:-1]))
            validator_fn = getattr(module, path_parts[-1])
            validator = validator_fn(self.graph, self.doc_lines, url=self.url)
            self.validators.add(validator)

    def _document_lines(self, text):
        """helper, get a list of (linetext, linenum) from a string with newlines"""
        inlines = text.split('\n')
        doc_lines = [(re.sub(r'^ +| +$', '', line), num)
                     for line, num
                     in zip(inlines, xrange(1, len(inlines) + 1))]
        return doc_lines

    def _get_document(self, source):
        """helper, open a file or url and return the content and identifier"""
        scheme_url = source
        if not source.startswith("http"):
            scheme_url = "http://%s" % source

        text = source

        try:
            text = urllib.urlopen(scheme_url).read()
        except:
            pass
        else:
            return (text, scheme_url)

        try:
            text = open(source, "r").read()
        except:
            pass
        else:
            return (text, source)

        return (text, None)



import re


def error_line(string, doc_lines=None):
    for line,num in doc_lines:
        if string in line:
            return line, num
            line = line.split(' ')
            for a in line:
                if string in a:
                    line = ' '.join(line[line.index(a) - 3 : line.index(a) + 3])
                    index = line.find(string)
                    line = line[index - 15 : index + len(string) + 15]
                    return line,num
    return ('', 0)

def _error(message, *strings, **kwargs):
    search_string = kwargs['search_string'] if 'search_string' in kwargs.keys() else strings[0]
    doc_lines = kwargs['doc_lines']
    message = message.format(*strings)
    line,num = error_line(search_string, doc_lines=doc_lines)
    return {'num': num, 'line': line, 'err': message}



from validator import MicrodataValidator
from schemadef import MicrodataSchemaDef


class SchemaOrgSchemaDef(MicrodataSchemaDef):
    def __init__(self):
        super(SchemaOrgSchemaDef, self).__init__()
        self._ontology_file = "http://schema.org/docs/schema_org_rdfa.html"
        self._representation = "schemaorg_schemadef"


class SchemaOrgValidator(MicrodataValidator):
    def __init__(self, graph, doc_lines, url=""):
        super(SchemaOrgValidator, self).__init__(graph, doc_lines, url=url)
        self.schema_def = SchemaOrgSchemaDef()
        self.allowed_namespaces = ["http://schema.org/"]



import json


class ValidationResult(object):
    ERROR = 1
    WARNING = 2

    def __init__(self, namespace, classname):
        super(ValidationResult, self).__init__()
        self.warnings = []
        self.errors = []
        self.namespace = namespace
        self.classname = classname

    def add_error(self, warning):
        if warning:
            if warning.level == ValidationResult.WARNING:
                self.warnings.append(warning)
            elif warning.level == ValidationResult.ERROR:
                self.errors.append(warning)

    def to_json(self):
        mapping = self.to_dict()
        return json.dumps(mapping)

    def to_dict(self):
        mapping = {}
        mapping['warnings'] = []
        for warning in self.warnings:
            mapping['warnings'].append(warning.to_dict())
        mapping['errors'] = []
        for error in self.errors:
            mapping['errors'].append(error.to_dict())
        mapping['namespace'] = self.namespace
        mapping['classname'] = self.classname
        return mapping

    def __len__(self):
        return len(self.warnings) + len(self.errors)


class ValidationWarning(object):
    def __init__(self, level, string, line, line_num):
        super(ValidationWarning, self).__init__()
        self.level = level
        self.string = string
        self.line_num = line_num
        self.line_text = line

    def to_dict(self):
        mapping = {}
        mapping['level'] = "Error" if self.level == ValidationResult.ERROR else "Warning"
        mapping['string'] = self.string
        mapping['line'] = self.line_text
        mapping['num'] = self.line_num
        return mapping

    def to_json(self):
        mapping = self.to_dict()
        return json.dumps(mapping)



from distillers import Distill, Distiller


class NewsDistiller(Distiller):
    site = Distill("og:site_name")
    title = Distill("s:headline", "og:title")
    image_url = Distill("s:associatedMedia.ImageObject/url", "og:image")
    pub_date = Distill("s:datePublished")
    author = Distill("s:creator.Person/name", "s:author")
    section = Distill("s:articleSection")
    description = Distill("s:description", "og:description")
    link = Distill("s:url", "og:url")
    id = Distill("s:identifier")


class ParselyDistiller(Distiller):
    site = Distill("og:site_name")
    title = Distill("pp:title", "s:headline", "og:title")
    image_url = Distill("pp:image_url", "s:associatedMedia.ImageObject/url", "og:image")
    pub_date = Distill("pp:pub_date", "s:datePublished")
    author = Distill("pp:author", "s:creator.Person/name", "s:author")
    section = Distill("pp:section", "s:articleSection")
    link = Distill("pp:link", "og:url", "s:url")
    post_id = Distill("pp:post_id", "s:identifier")
    page_type = Distill("pp:type")



from setuptools import setup

reqs = ['rdflib',
        'lepl',
        'lxml'
        ]

setup(
    name="schemato",
    version="1.2.0",
    author='Emmett Butler',
    author_email='emmett@parsely.com',
    url='https://github.com/Parsely/schemato',
    keywords='microdata rdf metadata',
    packages=['schemato', 'schemato.schemas'],
    description='Unified semantic metadata validator',
    platforms='any',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    install_requires=reqs
)



from validator import RdfValidator
from schemadef import RdfSchemaDef


class RNewsValidator(RdfValidator):
    def __init__(self, graph, doc_lines, url=""):
        super(RNewsValidator, self).__init__(graph, doc_lines, url=url)
        self.schema_def = RNewsSchemaDef()
        self.allowed_namespaces = ["http://iptc.org/std/rNews/2011-10-07#"]


class RNewsSchemaDef(RdfSchemaDef):
    def __init__(self):
        super(RNewsSchemaDef, self).__init__()
        self._ontology_file = "http://dev.iptc.org/files/rNews/rnews_1.0_draft3_rdfxml.owl"
        self._representation = "rnews_schemadef"



from rdflib.term import URIRef


class Distill(object):
    def __init__(self, *precedence, **kwargs):
        self.precedence = precedence

    def __repr__(self):
        return "Distill(name={0}, precedence={1})".format(self.name, self.precedence)


class DistillerMeta(type):
    def __new__(meta, classname, bases, class_dict):
        distill_fields = []
        for name, value in class_dict.iteritems():
            if isinstance(value, Distill):
                if getattr(value, "name", None) is None:
                    value.name = name
                distill_fields.append(value)
        class_dict["distill_fields"] = distill_fields
        return type.__new__(meta, classname, bases, class_dict)


class Distiller(object):
    __metaclass__ = DistillerMeta

    def __init__(self, schemato):
        self.schemato = schemato
        self.distilled = {}
        self.sources = {}
        # self.distill_fields provided by DistillerMeta

    def distill(self):
        for field in self.distill_fields:
            value, path = self.get_field(field)
            self.distilled[field.name] = value
            self.sources[field.name] = path
        return self.distilled

    def get_field(self, field):
        for path in field.precedence:
            value = self.get_value(path)
            if value is not None:
                return value, path
        return (None, None)

    PATH_PREFIX = {
        "pp": "get_parsely_page",
        "s": "get_schema_org",
        "og": "get_open_graph"
    }

    def parse_path(self, path):
        prefix, field = path.split(":")
        segments = field.split(".")
        return prefix, segments

    def get_parsely_page(self, segments):
        # for pp, we don't support nesting, so there will always
        # be a single segment
        assert len(segments) == 1
        key = segments[0]
        pp = self.schemato.parsely_page
        if pp is not None and key in pp:
            return pp[key]

    def get_open_graph(self, segments):
        # for opengraph, no nesting
        assert len(segments) == 1
        key = segments[0]
        graph = self.schemato.graph.rdfa_graph
        matches = graph.triples(
            (None, URIRef("http://ogp.me/ns#{key}".format(key=key)), None)
        )
        for match in matches:
            subj, pred, obj = match
            return ''.join((obj,)).encode('utf-8')
        return None

    def get_schema_org(self, segments):
        graph = self.schemato.graph.microdata_graph
        if len(segments) == 1:
            segment = segments[0]
            matches = graph.triples(
                (None, URIRef("http://schema.org/{key}".format(key=segment)), None)
            )
            match = None
            try:
                match = matches.next()
            except StopIteration:
                pass
            if match is None:
                return None
            subj, pred, obj = match
            return ''.join((obj,)).encode('utf-8')
        elif len(segments) == 2:
            root, nested = segments
            discriminator, field = nested.split("/")
            typematches = graph.triples(
                (None, URIRef("http://www.w3.org/1999/02/22-rdf-syntax-ns#type"), URIRef("http://schema.org/{key}".format(key=discriminator)))
            )
            if field == "url":
                for match in typematches:
                    subj, pred, obj = match
                    if isinstance(subj, URIRef):
                        return ''.join((subj,)).encode('utf-8')
                return None
            else:
                typematch = None
                try:
                    typematch = typematches.next()
                except StopIteration:
                    typematch = (None, None, None)
                subj, pred, obj = typematch
                if subj is None:
                    return None
                valmatches = graph.triples(
                    (subj, URIRef("http://schema.org/{key}".format(key=field)), None)
                )
                valmatch = None
                try:
                    valmatch = valmatches.next()
                except StopIteration:
                    valmatch = (None, None, None)
                subj, pred, obj = valmatch
                if obj is None:
                    return None
                return ''.join((obj,)).encode('utf-8')
        else:
            raise NotImplemented("Only 1 or 2 path segments currently supported")

    def get_value(self, path):
        prefix, segments = self.parse_path(path)
        i = 0
        method_name = self.PATH_PREFIX[prefix]
        method = getattr(self, method_name)
        for segment in segments:
            print " " * i, prefix, segment
            i += 1
        ret = method(segments)
        if ret is not None:
            print " " * i + " -> " + ret
        return ret
